/*
 * pktgen.c
 *
 * libnet demostration code for Webb.it 2004 Padova
 * coded by: Marco Balduzzi <embyte@spine-group.org>
 * compile with: gcc -Wall -O2 pktforger.c -o pktforger -lnet
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; see
 *   http://www.opensource.org/licenses/gpl-license.html for more info.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT *ANY KIND OF WARRANTY*
 *
 */

#include <libnet.h>

void give_help()
{
   printf ("pktgen.c\n");
   printf ("libnet demostration code for Webb.it 2004 Padova\n");
   printf ("licensed under GPL by <embyte@spine-group.org>\n");
   printf ("\n");

   printf ("usage: pktgen\n");
   printf ("-i  icmp ping\n");
   printf ("-u  udp packet\n");
   printf ("-s  source address (default: autodetected)\n");
   printf ("-d  destination address\n");
   printf ("-p  destination port\n");
   printf ("-m  udp message (default: none)\n");
   printf ("-h  print help\n");

   exit (-1);
}

int main(int argc, char **argv)
{
   struct input_struct
     {
	u_short type; 		/* 0 for icmp, 1 for udp, 2 error */
        u_long s_ip;   		/* source ip address */
	u_long d_ip;   		/* destination ip address */
	u_short d_port; 	/* destination port */
        u_short s_port;		/* source port */
	u_char *data;		/* udp payload */
	u_short data_size;	/* udp payload size */
     }
   input;

   u_short h_lenght;		/* header lenght */
   u_short ip_proto;		/* ip-proto ip header field */
   short n;			/* byte wrote */
   int option;			/* getopt */
   char err_buf[LIBNET_ERRBUF_SIZE]; /* error buffer */

   libnet_t *l;			/* libnet context handler */

   /* set default */
   input.type=2;
   input.d_ip=0;
   input.d_port=0;
   input.s_port=10000;
   input.data_size=0;
   input.data=NULL;
   
   /* initialize libnet_t* */
   if ((l=libnet_init(LIBNET_RAW4, NULL, err_buf))==NULL)
     {
	fprintf (stderr, "Error: %s\n", err_buf);
	return -1;
     }

   /* autodetect iface ip */
   if ((input.s_ip = libnet_get_ipaddr4(l))==-1)
     {
	fprintf (stderr, "Error: %s\n", libnet_geterror(l));
	return -1;
     }

   /* parse cmdline */
   while ((option = getopt (argc, argv, "uis:d:p:m:hH?"))!=EOF)
     {
	switch (option)
	  {
	   case 'u':
	     input.type=1;
	     break;
	   case 'i':
	     input.type=0;
	     break;
	   case 's':
	     if ((input.s_ip = libnet_name2addr4(l, optarg, LIBNET_RESOLVE))==-1)
	       {
		  fprintf (stderr, "Error: %s\n", libnet_geterror(l));
		  return -1;
	       }
	     break;
	   case 'd':
	     if ((input.d_ip = libnet_name2addr4(l, optarg, LIBNET_RESOLVE))==-1)
	       {
		  fprintf (stderr, "Error: %s\n", libnet_geterror(l));
		  return -1;
	       }
	     break;
	   case 'p':
	     input.d_port = (u_short) atoi(optarg);
	     break;
	   case 'm':
	     (input.data = (u_char *) optarg);
	     input.data_size = strlen ((char *) input.data);
	     break;
	   case 'h':
	   case 'H':
	   case '?':
	   default:
	     give_help();
	     break;
	  }
     }

   /* some input checks */
   if (input.type==2)
     {
	fprintf (stderr, "Error: you must specify icmp or udp packet type (-i or -u)\n\n");
	give_help();
	return -1;
     }

   if (!input.d_ip)
     {
	fprintf (stderr, "Error: you must specify a destination address\n\n");
	give_help();
	return -1;
     }

   if (!input.d_port && input.type==1)
     {
	fprintf (stderr, "Error: you must specify a destination port\n\n");
	give_help();
	return -1;
     }

   if (input.data_size && !input.type)
     fprintf (stderr, "* Ignoring payload (icmp)\n");

   if (!input.type)	/* ICMP */
     {
        /* build icmp echo-request header */
	if (libnet_build_icmpv4_echo
	    (
	     ICMP_ECHO,	/* type */
	     0,		/* code */
	     0,		/* sum */
	     0x100,	/* id */
	     0x100,	/* seq */
	     NULL,	/* payload */
	     0,		/* payload lenght */
	     l,		/* libnet context pointer */
	     0 		/* We don't need ptag */
	     )==-1)
	  {
	     fprintf (stderr, "Error %s\n", libnet_geterror(l));
	     return 1;
	  }

	/* options for ip layer */
	h_lenght=LIBNET_ICMPV4_ECHO_H+LIBNET_IPV4_H;
	ip_proto=IPPROTO_ICMP;

	printf ("* ICMP header builded\n");
     }
   else			/* UDP */
     {
	/* build icmp udp header */
	if (libnet_build_udp
	    (
	     input.s_port,	/* source port */
	     input.d_port,	/* destination port */
	     LIBNET_UDP_H + input.data_size,	/* size */
	     0,			/* sum */
	     input.data,	/* payload */
	     input.data_size,	/* payload size */
	     l,			/* libnet context */
	     0			/* we don't need ptag */
	     )==-1)
	  {
	     fprintf (stderr, "Error %s\n", libnet_geterror(l));
	     return 1;
	  }

	/* options for ip layer */
	h_lenght=LIBNET_UDP_H+LIBNET_IPV4_H;
	ip_proto=IPPROTO_UDP;

	printf ("* UDP header builded\n");
     }

   /* build ipv4 header */
   if (libnet_build_ipv4
       (
	h_lenght+input.data_size,/* len */
	0,			/* tos */
	0x200,			/* id */
	0,			/* frag */
	64,			/* ttl */
	ip_proto,		/* transport protocol */
	0,			/* sum */
	input.s_ip,		/* source address */
	input.d_ip,		/* destination address */
	NULL,			/* ip payload */
	0,			/* ip payload size */
	l,			/* libnet context */
	0			/* we don't need ptag */
	)==-1)
     {

	fprintf (stderr, "Error %s\n", libnet_geterror(l));
	return 1;
     }

   printf ("* IP header builded\n");

   /* write packet */
   if ((n=libnet_write(l))==-1)
     fprintf (stderr, "Error %s\n", libnet_geterror(l));

   printf ("* Packet correctly wrote on wire: %d bytes\n", n);

   /* destroy all */
   libnet_destroy(l);

   return 0;
}

