/*
 * sniffer.c
 *
 * libpcap demostration code for Webb.it 2004 Padova
 * coded by: Paolo Carpo <snifth@spine-group.org>
 * compile with: gcc -Wall -O2 sniffer.c -o sniffer -lpcap
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; see
 *   http://www.opensource.org/licenses/gpl-license.html for more info.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT *ANY KIND OF WARRANTY*
 *
 */

/* includes */
#include <pcap.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdlib.h>
#include <signal.h>

/* prototipes of functions */
void dump_udp (char *packet, int offset);
void dump_icmp(char *packet, int offset);
void dump_generic(char *packet, int offset);
void sigexit();
void give_help(void);

pcap_t* descr;		/* our file descriptor for all pcap funcions */
			/* we use it as global var to use signal()   */

/* Define header protocol lenghts */
#define ICMPV4_H         0x04    /**< ICMP header base:     4 bytes */
#define IPV4_H           0x14    /**< IPv4 header:         20 bytes */
#define UDP_H            0x08    /**< UDP header:           8 bytes */

/* Set promiscuous mode of my NIC */
#define PROMISC          1

/*
 *  IPv4 header
 *  Internet Protocol, version 4
 *  Static header size: 20 bytes
 */
struct ipv4_hdr
{

   u_int8_t ip_hl:4,      /* header length */
     ip_v:4;         /* version */
   u_int8_t ip_tos;       /* type of service */
   u_int16_t ip_len;         /* total length */
   u_int16_t ip_id;          /* identification */
   u_int16_t ip_off;

   u_int8_t ip_ttl;          /* time to live */
   u_int8_t ip_p;            /* protocol */
   u_int16_t ip_sum;         /* checksum */
   struct in_addr ip_src, ip_dst; /* source and dest address */
};

/*
 *  ICMP header
 *  Internet Control Message Protocol
 *  Base header size: 4 bytes
 */
struct icmpv4_hdr
{
   u_int8_t icmp_type;       /* ICMP type */
#define     ICMP_ECHOREPLY                  0
#define     ICMP_UNREACH                    3
#define     ICMP_SOURCEQUENCH               4
#define     ICMP_REDIRECT                   5
#define     ICMP_ECHO                       8
#define     ICMP_ROUTERADVERT               9
#define     ICMP_ROUTERSOLICIT              10
#define     ICMP_TIMXCEED                   11
#define     ICMP_PARAMPROB                  12
#define     ICMP_TSTAMP                     13
#define     ICMP_TSTAMPREPLY                14
#define     ICMP_IREQ                       15
#define     ICMP_IREQREPLY                  16
#define     ICMP_MASKREQ                    17
#define     ICMP_MASKREPLY                  18

   u_int8_t icmp_code;       /* ICMP code */

#define     ICMP_UNREACH_NET                0
#define     ICMP_UNREACH_HOST               1
#define     ICMP_UNREACH_PROTOCOL           2
#define     ICMP_UNREACH_PORT               3
#define     ICMP_UNREACH_NEEDFRAG           4
#define     ICMP_UNREACH_SRCFAIL            5
#define     ICMP_UNREACH_NET_UNKNOWN        6
#define     ICMP_UNREACH_HOST_UNKNOWN       7
#define     ICMP_UNREACH_ISOLATED           8
#define     ICMP_UNREACH_NET_PROHIB         9
#define     ICMP_UNREACH_HOST_PROHIB        10
#define     ICMP_UNREACH_TOSNET             11
#define     ICMP_UNREACH_TOSHOST            12
#define     ICMP_UNREACH_FILTER_PROHIB      13
#define     ICMP_UNREACH_HOST_PRECEDENCE    14
#define     ICMP_UNREACH_PRECEDENCE_CUTOFF  15
#define     ICMP_REDIRECT_NET               0
#define     ICMP_REDIRECT_HOST              1
#define     ICMP_REDIRECT_TOSNET            2
#define     ICMP_REDIRECT_TOSHOST           3
#define     ICMP_TIMXCEED_INTRANS           0
#define     ICMP_TIMXCEED_REASS             1
#define     ICMP_PARAMPROB_OPTABSENT        1

   u_int16_t icmp_sum;   /* ICMP Checksum */

   union
     {
        struct
	  {
	     u_int16_t id; /* ICMP id */
	     u_int16_t seq;/* ICMP sequence number */
	  }
	echo;

#undef icmp_id
#undef icmp_seq
#define icmp_id     hun.echo.id
#define icmp_seq    hun.echo.seq

        u_int32_t gateway; /* gateway host */
        struct
	  {
	     u_int16_t pad;/* padding */
	     u_int16_t mtu;/* MTU size */
	  }
	frag;
     }
   hun;
};

/*
 *  UDP header
 *  User Data Protocol
 *  Static header size: 8 bytes
 */
struct udp_hdr
{
   u_int16_t uh_sport;       /* soure port */
   u_int16_t uh_dport;       /* destination port */
   u_int16_t uh_ulen;        /* length */
   u_int16_t uh_sum;         /* checksum */
};

void give_help()
{
   printf ("\nsniffer.c\n");
   printf ("libpcap demostration code for Webb.it 2004 Padova\n");
   printf ("licensed under GPL by <snifth@spine-group.org>\n");
   printf ("\n");

   printf ("usage: sniffer\n");
   printf ("-i  interface\n");
   printf ("-f  \"filter\"\n");
   printf ("-h  print help\n");

   exit (-1);
}

int main(int argc, char **argv)
{
   bpf_u_int32 maskp;		/* netmask */
   bpf_u_int32 netp;
   struct bpf_program fp;	/* pointer to bpf_program struct */
   				/* filled in by pcap_compile() with our filter */
   struct pcap_pkthdr hdr;	/* struct filled in by pcap_next() */

   struct ipv4_hdr *ip;		/* pointer to ipv4 header */

   int offset;			/* offset to jump for read protocol headers */
   char *packet;		/* our packet */
   int datalink;		/* type of datalink we use */
   char *dev;			/* our NIC name */
   char *filter;		/* the filter we want to apply */

   int option;			/* options by the shell (horrible english)*/

   char errbuf[PCAP_ERRBUF_SIZE]; /* error buffer */

   dev = filter = packet = NULL;

   /* parse cmdline */
   while ((option = getopt (argc, argv, "i:f:H?"))!=EOF)
     {
	switch (option)
	  {
	   case 'i':
	     dev = optarg;
	     break;
	   case 'f':
	     filter = optarg;
	     break;
	   case 'h':
	   case 'H':
	   case '?':
	   default:
	     give_help();
	     break;
	  }
     }

   /* if we don't give a device to use...exit */
   if(dev == NULL)
     {
	fprintf (stderr, "Which interface?\n");
	give_help();
     }

   /* signal handlers */
   signal(SIGKILL, sigexit);
   signal(SIGQUIT, sigexit);
   signal(SIGTERM, sigexit);
   signal(SIGINT, sigexit);

   /* check if the dev is up and read its parameters */
   if ((pcap_lookupnet(dev,&netp,&maskp,errbuf))==-1)
     {
	fprintf (stderr, "Error: pcap_lookupnet(): %s\n\n", errbuf);
	return -1;
     }

   /* open our interface in promisc mode and without timeout */
   if ((descr = pcap_open_live (dev, BUFSIZ, PROMISC, 0, errbuf))==NULL)
     {
	fprintf(stderr, "Error: pcap_open_live(): %s\n\n",errbuf);
	return -1;
     }

   /* look for our datalink and get our offset */
   if ((datalink = pcap_datalink(descr)) < 0)
     {
	fprintf(stderr, "Error: pcap_datalink(): %s\n\n", errbuf);
	return -1;
     }
   switch (datalink)
     {
      case DLT_EN10MB:
	offset = 14;
	break;
      case DLT_NULL:
      case DLT_PPP:
	offset = 4;
	break;
      case DLT_SLIP:
	offset = 16;
	break;
      case DLT_RAW:
	offset = 0;
	break;
      case DLT_SLIP_BSDOS:
      case DLT_PPP_BSDOS:
	offset = 24;
	break;
      case DLT_FDDI:
	offset = 21;
	break;
      default:
	fprintf(stderr, "Error: Unknown Datalink Type: (%d)\n\n", datalink);
	return -1;
     }

   /* compile the filter */
   if(pcap_compile(descr,&fp,filter,0,netp) == -1)
     {
	fprintf(stderr, "Error calling pcap_compile()\n\n");
	return -1;
     }

   /* set the filter */
   if(pcap_setfilter(descr,&fp) == -1)
     {
	fprintf(stderr, "Error calling pcap_setfilter()\n\n");
	return -1;
     }

   /* read all the packets searching for UDP & ICMP */
   while(1)
     {

     	/* read with pcap_next() */
	if ((packet = (char *) pcap_next (descr, &hdr)) == NULL) continue;
	ip = (struct ipv4_hdr *) (packet+offset);

	switch(ip->ip_p)
	  {
	  /*if the packet is UDP, print its options */
	   case IPPROTO_UDP:
	     dump_udp(packet, offset);
	     break;
	   /*if the packet is ICMP, print its options */
	   case IPPROTO_ICMP:
	     dump_icmp(packet, offset);
	     break;
	   /*print generic packets */
	   case IPPROTO_TCP:
	   case IPPROTO_IGMP:
	     dump_generic(packet, offset);
	     break;
	  }

     }
   return 0;
}

void dump_udp(char *packet, int offset)
{
   struct ipv4_hdr *ip;
   struct udp_hdr *udp;
   struct servent *service;
   u_short size_buf, size_ip, size_udp;
   char *buf;

   size_ip = IPV4_H;
   size_udp = UDP_H;
   size_buf = 0;
   buf = NULL;

   ip = (struct ipv4_hdr *) (packet+offset);
   udp = (struct udp_hdr *) (packet+size_ip+offset);

   size_buf = ntohs(ip->ip_len) - size_ip - size_udp;

   printf("\n---[ UDP ]--------------------------------------------\n");
   service = getservbyport (htons(ntohs(udp->uh_sport)), "udp");
   printf("%s:%d(%s)", inet_ntoa (ip->ip_src), ntohs(udp->uh_sport), (service) ? service->s_name : "unknown");

   service = getservbyport(htons(ntohs(udp->uh_dport)), "udp");
   printf(" -> ");
   printf("%s:%d(%s)\n", inet_ntoa(ip->ip_dst), ntohs(udp->uh_dport), (service) ? service->s_name : "unknown");

   printf("Version: %d\t", ip->ip_v);
   printf("Lenght IP: %d\t", ntohs(ip->ip_len));
   printf("Lenght UDP: %d\t",ntohs(udp->uh_ulen));
   printf("TTL: %d\n", ip->ip_ttl);
   /* retrive payload */
   if (size_buf)
     {
	buf = (char *) (packet + size_ip + size_udp + offset);
	printf("Payload lenght: %d\t",size_buf);
	printf("Payload: %s\n",buf);
     }
}

void dump_icmp(char *packet, int offset)
{
   struct ipv4_hdr *ip;
   struct icmpv4_hdr *icmp;
   u_short size_ip, size_icmp, size_buf;

   size_ip = IPV4_H;
   size_icmp = ICMPV4_H; /* base ICMP header lenght */
   size_buf = 0;

   ip = (struct ipv4_hdr *) (packet+offset);
   icmp = (struct icmpv4_hdr *) (packet+size_ip+offset);

   printf("\n---[ ICMP ]--------------------------------------------\n");
   printf("%s", inet_ntoa(ip->ip_src));
   printf(" -> ");
   printf("%s\n", inet_ntoa(ip->ip_dst));
   printf("Version: %d\t", ip->ip_v);
   printf("Lenght IP: %d\t", ntohs(ip->ip_len));
   printf("TTL: %d\n", ip->ip_ttl);
   printf("Type: ");

   switch((icmp->icmp_type))
     {
      case 0:
	printf("Echo reply\n");
	size_icmp+=4;
	break;
      case 3:
	printf("Dest. unreach: ");
	size_icmp+=4;
	switch (icmp->icmp_code)
	  {
	   case 0:
	     printf("Network Unreachable\n");
	     break;
	   case 1:
	     printf("Host Unreachable\n");
	     break;
	   case 2:
	     printf("Protocol Unreachable\n");
	     break;
	   case 3:
	     printf("Port Unreachable\n");
	     break;
	   case 4:
	     printf("Fragmentation neded (DF)\n");
	     break;
	   case 5:
	     printf("Source route failed\n");
	     break;
	   case 6:
	     printf("Destination network unknown\n");
	     break;
	   case 7:
	     printf("Destination host unknown\n");
	     break;
	   case 8:
	     printf("Source host isolated\n");
	     break;
	   case 9:
	     printf("Destination network administratively prohibited\n");
	     break;
	   case 10:
	     printf("Destination host administratively prohibited\n");
	     break;
	   case 11:
	     printf("Network unreacjable(tOS)\n");
	     break;
	   case 12:
	     printf("Host Unreachable (tOS)\n");
	     break;
	   case 13:
	     printf("Communication administratively prohibited\n");
	     break;
	   case 14:
	     printf("Host precedence violation\n");
	     break;
	   case 15:
	     printf("Precedence cutoff in effect\n");
	     break;
	   default:
	     printf("Unknown - error?\n");
	     break;
	  }
	break;
      case 4:
	printf("Source quench\n");
	size_icmp+=4;
	break;
      case 5:
	printf("Redirect: ");
	size_icmp+=4;
	switch(icmp->icmp_code)
	  {
	   case 0:
	     printf("Redirect for network\n");
	     break;
	   case 1:
	     printf("Redirect for host\n");
	     break;
	   case 2:
	     printf("Redircet for tos & network\n");
	     break;
	   case 3:
	     printf("Redirect for tos & host\n");
	     break;
	   default:
	     printf("Unknown - error?\n");
	     break;
	  }
	break;
      case 8:
	printf("Echo request\n");
	size_icmp+=4;
	break;
      case 11:
	printf("Time exceeded: ");
	size_icmp+=4;
	switch (icmp->icmp_code)
	  {
	   case 0:
	     printf("TTL (0) during transit\n");
	     break;
	   case 1:
	     printf("TTL (0) during reassembly\n");
	     break;
	   default:
	     printf("Unknown - error?\n");
	     break;
	  }
	break;
      case 12:
	printf("Parameter problem: ");
	switch (icmp->icmp_code)
	  {
	   case 0:
	     printf("IP header bad\n");
	     break;
	   case 1:
	     printf("Requiring option missing\n");
	     break;
	  }
	break;
      case 13:
	printf("Timestamp\n");
	size_icmp+=16;
	break;
      case 14:
	printf("Timestamp reply\n");
	size_icmp+=16;
	break;
      case 15:
	printf("Information\n");
	break;
      case 16:
	printf("Information reply\n");
	break;
      case 17:
	printf("Address mask\n");
	size_icmp+=8;
	break;
      case 18:
	printf("Address mask reply\n");
	size_icmp+=8;
	break;
      default:
	printf("%i\n", icmp->icmp_type);
	break;
     }
}

void dump_generic(char *packet, int offset)
{
   struct ipv4_hdr *ip = (struct ipv4_hdr *) (packet+offset);

   printf("\n---[ IP packet ]-----------------------------------------\n");
   printf("%s", inet_ntoa(ip->ip_src));
   printf(" -> ");
   printf("%s\n", inet_ntoa(ip->ip_dst));
   printf("Version: %d\t", ip->ip_v);
   printf("Lenght IP: %d\t", ntohs(ip->ip_len));
   printf("TTL: %d\n", ip->ip_ttl);
   printf("Transport protocol number: %d\n", ip->ip_p);
}

/* signal handler */
void sigexit()
{
   struct pcap_stat statistic;

   if (pcap_stats(descr,&statistic) < 0)
     fprintf(stderr, "Error: pcap_stats: %s\n", pcap_geterr(descr));
   else
     {
	printf("\nPackets Received:\t\t%d\n", statistic.ps_recv);
	printf("Packets Dropped by kernel:\t%d\n", statistic.ps_drop);
     }

   pcap_close(descr);
   exit(0);
}
